//PRODUCTION APP Version 0.5.1
#include "globals.h"
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <SPI.h>
#include "WiFi.h"
#include "Audio.h"
#include <ezButton.h>
#include "display.h"
#include "audiostatus.h"
// Update for secure connection
#include <WiFiClientSecure.h>
// Update for Setup Mode Web Server
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <SPIFFS.h>
#include "wifisetup.h"



//*************************************************************************
// Define I2S connections for Sparkfun MAX Audio Amplifier/DAC
//*************************************************************************
#define I2S_DOUT  5   //DIN on MAX Board
#define I2S_BCLK  26  //BCLK
#define I2S_LRC   25  //LRCLK on MAX Board
// Create audio object
Audio audio;

//Create variables to hold songtitle and track if it changed
String songTitle = "";  //Used for displaying on screen
String prevsongTitle = "";
bool songTitleChanged = false;

//Variables to clear volume display
unsigned long startTime = 0; // Variable to store the start time
bool displayVolume = false; // Flag to indicate whether to display volume

//*************************************************************************
// Define which line number each text element will display on
// For text size of 1:  Line should be 1-29
// For text size of 2: Line should be 1-8
//*************************************************************************
#define SONGTITLE_SIZE  1
#define SONGTITLE_LINE  28
#define MAINTITLE_SIZE  2
#define MAINTITLE_LINE  4
#define SUBTITLE_SIZE   1
#define SUBTITLE_LINE   20
#define CATEGORY_SIZE   2
#define CATEGORY_LINE   6
#define STATION_SIZE    2
#define STATION_LINE    6
#define PROMPT_SIZE     2
#define PROMPT_LINE     3
#define VOLUME_SIZE     1
#define VOLUME_LINE     25
#define VOLUME_DISPLAY_TIME 3500  //Time is in ms

//*************************************************************************
// Setup variables for Categories & Stations
//*************************************************************************
const int MAX_CATEGORIES = 20; // Maximum number of categories
const int MAX_STATIONS_PER_CATEGORY = 25; // Maximum number of stations per category

String categories[MAX_CATEGORIES]; // Array to store category names
String stations[MAX_CATEGORIES][MAX_STATIONS_PER_CATEGORY]; // 2D array to store stations per category

int categoryIndex = -1; // Current index for categories array
int stationIndex = 0; // Current index for stations array

String stationName = "";
String stationUrl = "";

//*************************************************************************
// Setup defines and variables to track state for menu display and selection
//*************************************************************************
#define VOLUME 1    //Used for storing the state
#define CATEGORY 2  //Used for storing the state
#define STATION 3   //Used for storing the state
int currentCategory = 0;  
int currentStation = 0;
int volume = 5;
int state = VOLUME;
bool skipButton = false;

//*************************************************************************
//Pins for display & setup for display
//*************************************************************************
#define TFT_DC 21
#define TFT_CS 22 
#define TFT_MOSI 23
#define TFT_CLK 18
#define TFT_RST 17
#define TFT_MISO 19

// Use hardware SPI (on Uno, #13, #12, #11) and the above for CS/DC
//Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC);
// If using the breakout, change pins as desired
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);

//*************************************************************************
//Pins for rotary encoder and variables
//*************************************************************************
// Rotary Encoder Inputs
 #define inputCLK 34
 #define inputDT 35
 #define inputSW 32
 #define DIRECTION_CW 1
 #define DIRECTION_CCW 2
 int counter = 0; 
 int currentStateCLK;
 int previousStateCLK; 
 ezButton button(inputSW);
 String encdir ="";
 int direction;

 // Setup vars to be used for wifi and host
String ssidValue = "";
String passwordValue = "";
String hostValue = "";
String pathValue = "";


//*************************************************************************
// Setup
//*************************************************************************
void setup() {
  // Initialize the serial output
  //Serial.begin(115200);


  // Connect MAX98357 I2S Apm, Set initial volume & Timeout parms
  audio.setPinout(I2S_BCLK, I2S_LRC, I2S_DOUT);
  audio.setVolume(volume);  //Set 0 - 100
  audio.setConnectionTimeout(1000, 1000);

  // configure encoder pins as inputs
  // Set encoder pins as inputs  
  pinMode (inputCLK,INPUT);
  pinMode (inputDT,INPUT);
  pinMode (inputSW,INPUT_PULLUP);
  button.setDebounceTime(30); 

  // Initialize the display
  tft.begin();
  tft.setRotation(1);
  tft.fillScreen(ILI9341_BLACK);
  tft.setTextColor(ILI9341_WHITE);

  //Initialise SPIFFS
  if(!SPIFFS.begin(true)){
      //Serial.println("SPIFFS Mount Failed");
      return;
  }
// Now that SPIFFS is working, read the values 
ssidValue = readFile(SPIFFS, "/ssid");
passwordValue = readFile(SPIFFS, "/password");
hostValue = readFile(SPIFFS, "/host");
pathValue = readFile(SPIFFS, "/path");

// DECIDE if we need to go to SETUP mode or not.
// Check if any value is not set or if the user holds down the button while turning on.
if (ssidValue.isEmpty() || passwordValue.isEmpty() || hostValue.isEmpty() || pathValue.isEmpty() || digitalRead(inputSW) == LOW) {
  tft.setTextSize(MAINTITLE_SIZE);
  displayCenteredText(tft,"<< Setup Mode >>",true,PROMPT_LINE);
  tft.setTextSize(SUBTITLE_SIZE);
  displayCenteredText(tft,"Connect to SSID: iRadio",false,SUBTITLE_LINE-1);
  displayCenteredText(tft,"Open your browswer and navigate to", false,SUBTITLE_LINE+1);
  displayCenteredText(tft,"\"192.168.4.1\" to complete setup.", false,SUBTITLE_LINE+2);
  setupWiFi();
}



  // Setup and display title screen
  tft.setTextSize(MAINTITLE_SIZE);
  displayCenteredText(tft,"iRadio",true,PROMPT_LINE);
  tft.setTextSize(SUBTITLE_SIZE);
  displayCenteredText(tft,"by Chris Fulton",false,SUBTITLE_LINE);
  displayCenteredText(tft,"Initializing",false,SUBTITLE_LINE + 2);


  //Call function to establish wifi connetion
  //and to read the category/station list and put into the arrays
  wifi_and_station_setup();

  // Read the initial state of inputCLK
  // Assign to previousStateCLK variable
   previousStateCLK = digitalRead(inputCLK);


  //  Don't play on startup.
  //audio.connecttohost("www.surfmusic.de/m3u/100-5-das-hitradio,4529.m3u");
  
  
  // Erase title screen
  tft.setTextSize(MAINTITLE_SIZE);
  eraseCenteredText(tft,"iRadio",true,PROMPT_LINE);
  tft.setTextSize(SUBTITLE_SIZE);
  eraseCenteredText(tft,"by Chris Fulton",false,SUBTITLE_LINE);
  eraseCenteredText(tft,"Initializing",false,SUBTITLE_LINE + 2);

  //Set state to CATEGORY so user can select after booting.
  state = CATEGORY;
  tft.setTextSize(CATEGORY_SIZE);
  displayCenteredText(tft,"Select Category:",true,PROMPT_LINE); 
  displayCenteredText(tft,categories[currentCategory].c_str(),false,CATEGORY_LINE);
}



//*************************************************************************
// loop
//*************************************************************************
void loop() {


audio.loop();
button.loop();

//Check if volume needs cleared
if (displayVolume && (millis() - startTime >= VOLUME_DISPLAY_TIME)) {
    // Time elapsed, erase volume text
    eraseCenteredText(tft, String(volume).c_str(), false, VOLUME_LINE);
    displayVolume = false; // Reset flag
  }

//Has the song title changed, if so update the display.
//The stream will share this data and we capture it in audiostatus
if (songTitleChanged) {
  tft.setTextSize(SONGTITLE_SIZE);
  eraseCenteredText(tft, prevsongTitle.c_str(), false, SONGTITLE_LINE);
  displayCenteredText(tft, songTitle.c_str(), false, SONGTITLE_LINE);
  prevsongTitle = songTitle;
  songTitleChanged = false;
}

//Handle when the button is pressed/released.
//Would not work with ispressed for some reason
if (button.isReleased()) {
  // We will clear the volume display in case it is on the screen
  tft.setTextSize(VOLUME_SIZE);
  eraseCenteredText(tft, String(volume).c_str(), false, VOLUME_LINE);


    if (state == VOLUME) {
    state = CATEGORY;
    //Serial.println("State=Category");
    tft.setTextSize(CATEGORY_SIZE);
    eraseCenteredText(tft,stationName.c_str(),false,CATEGORY_LINE);
    displayCenteredText(tft,"Select Category:",true,PROMPT_LINE); 
    displayCenteredText(tft,categories[currentCategory].c_str(),false,CATEGORY_LINE);
  }
  else if (state == CATEGORY) {
    //Serial.println("State=STATION");
    state = STATION;
    currentStation=0;  //Added to fix overwrite screen issue??
    tft.setTextSize(STATION_SIZE); //yyyy
    eraseCenteredText(tft,"Select Category:", true, PROMPT_LINE);
    eraseCenteredText(tft,categories[currentCategory].c_str(),false,CATEGORY_LINE);
    displayCenteredText(tft,"Select Station:", true, PROMPT_LINE); 
    stationName = parseStation(stations[currentCategory][0],true);
    stationUrl = parseStation(stations[currentCategory][0],false);
    displayCenteredText(tft,stationName.c_str(), false, STATION_LINE); 
  }
  else if (state == STATION) {
    audio.connecttohost(stationUrl.c_str());
    state = VOLUME;
    tft.setTextSize(SONGTITLE_SIZE);
    eraseCenteredText(tft,songTitle.c_str(),false,SONGTITLE_LINE);
    tft.setTextSize(2);
    eraseCenteredText(tft,"Select Station:", true, PROMPT_LINE);
  }

}

// Check to see if the encoder was rotated.
  // Read the current state of inputCLK
   currentStateCLK = digitalRead(inputCLK);
   // If the previous and the current state of the inputCLK are different then a pulse has occured
   if (currentStateCLK != previousStateCLK){        
     // If the inputDT state is different than the inputCLK state then 
     // the encoder is rotating counterclockwise
     if (digitalRead(inputDT) != currentStateCLK) { 
       counter --;
       encdir ="CCW";
       direction = DIRECTION_CCW;
       //Serial.println("<----");
       handleEncoderRotate();       
     } else {
       // Encoder is rotating clockwise
       counter ++;
       encdir ="CW";
       direction = DIRECTION_CW;
       //Serial.println("--->");
       handleEncoderRotate(); 
     }
   } 
   // Update previousStateCLK with the current state
   previousStateCLK = currentStateCLK; 

}


//*************************************************************************
// handleEncoderPress:
//*************************************************************************
void handleEncoderPress() {
//Serial.println("  In handleEncoderPress");
  if (state == VOLUME) {
    //Serial.println("State=Category");
    tft.setTextSize(2);
    displayCenteredText(tft,"Select Category:",true,PROMPT_LINE);
    displayCenteredText(tft,categories[currentCategory].c_str(),false,CATEGORY_LINE);
  }
  else if (state == CATEGORY) {
    //Serial.println("State=STATION");
    state = STATION;
    tft.setTextSize(2);
    eraseCenteredText(tft,"Select Category:", true, PROMPT_LINE);
    eraseCenteredText(tft,categories[currentCategory].c_str(),false,CATEGORY_LINE);
    displayCenteredText(tft,"Select Station", true, PROMPT_LINE);
    stationName = parseStation(stations[currentCategory][0],true);
    displayCenteredText(tft,stationName.c_str(), false, STATION_LINE);
    
  }
}

//*************************************************************************
// handleEncoderRotate: 
//*************************************************************************
void handleEncoderRotate() {
  switch (state) {
    case CATEGORY:
      tft.setTextSize(2);
      eraseCenteredText(tft, categories[currentCategory].c_str(), false, CATEGORY_LINE);
      currentCategory = (currentCategory + (direction == DIRECTION_CW ? -1 : 1) + getNumberOfCategories()) % getNumberOfCategories();
      displayCenteredText(tft, categories[currentCategory].c_str(), false, CATEGORY_LINE);
      break;

    case STATION:
      tft.setTextSize(2);
      eraseCenteredText(tft, parseStation(stations[currentCategory][currentStation], true).c_str(), false, STATION_LINE);
      currentStation = (currentStation + (direction == DIRECTION_CW ? -1 : 1) + getNumberOfStations(currentCategory)) % getNumberOfStations(currentCategory);
      stationName = parseStation(stations[currentCategory][currentStation], true);
      stationUrl = parseStation(stations[currentCategory][currentStation], false);
      displayCenteredText(tft, stationName.c_str(), false, STATION_LINE);
      break;

    case VOLUME:
      // Clear volume display
      eraseCenteredText(tft, String(volume).c_str(), false, VOLUME_LINE);
      // Adjust volume
      if (direction == DIRECTION_CW) {
        // Increase volume by 2
        volume -= 1;
      } else if (direction == DIRECTION_CCW) {
        // Decrease volume by 2
        volume += 1;
      }
      if (volume < 0) { volume = 0; }
      if (volume > 35) { volume = 35; }
      // Set the volume
      audio.setVolume(volume);
      // Display the updated volume on the screen (assuming you have a display function for this)
      tft.setTextSize(VOLUME_SIZE);
      displayCenteredText(tft, String(volume).c_str(), false, VOLUME_LINE);
      displayVolume = true; // Set flag to true when volume changes
      startTime = millis(); // Record the start time
      break;

    default:
      // Handle unexpected state
      break;
  }
}

//*************************************************************************
//Pass the station name/url string to this function.
//if getName = True, the name will be returned.
//if getName = false, the URL will be returned.
//*************************************************************************
String parseStation(String input, bool getName) {
  // Find the position of '{' and '}'
  int start_index = input.indexOf('{');
  int end_index = input.indexOf('}');
  
  if (start_index != -1 && end_index != -1) {
    if (getName) {
      // Extract the first part (name)
      return input.substring(0, start_index);
    } else {
      // Extract the second part (URL)
      return input.substring(start_index + 1, end_index);
    }
  } else {
    return "Error: Unable to parse the input string.";
  }
}

//*************************************************************************
// Setup the wifi, connect and then get the station list & parse it
//*************************************************************************
void wifi_and_station_setup() {
  // Setup WiFi in Station mode
  WiFi.disconnect();
  WiFi.mode(WIFI_STA);
//  const char* ssid = ssidValue.c_str();
//  const char* pass = passwordValue.c_str();


const char* ssid = ssidValue.c_str();
const char* password = passwordValue.c_str();

  WiFi.begin(ssid, password);
  
  while (WiFi.status() != WL_CONNECTED) {
    //Serial.print("SSID & PW " + ssidValue);
    //Serial.println(" passwordValue " + passwordValue);
    delay(2000);
    //Serial.println("Connecting to WiFi...");
  }
  //Serial.println("Connected to WiFi");

  // WiFi Connected, print IP to serial monitor
  //Serial.println("");
  //Serial.println("WiFi connected");
  //Serial.println("IP address: ");
  //Serial.println(WiFi.localIP());
  //Serial.println("");

  // Setup a secure client & disable cert validation for testing
  WiFiClientSecure client;
  client.setInsecure();

  //Serial.println();
  //Serial.print("Connecting to wifi: ");
  //Serial.println(ssidValue);
  //Serial.print("Path: ");
  //Serial.print(pathValue);

  // Connect to the serveer first.
  if (client.connect(hostValue.c_str(), 443)) {
    //Serial.println("Connected to server");

    // Then do a GET request for the specific text file
    client.print(String("GET ") + pathValue + " HTTP/1.1\r\n" +
             "Host: " + hostValue + "\r\n" +
             "Connection: close\r\n\r\n");
  
   // Read response headers
    String line;
    while (client.connected()) {
      line = client.readStringUntil('\n');
      // Check for end of headers (blank line)
      if (line == "\r") {
        break; // End of headers, exit the loop
      }
    }

    // Read response content (text file)
    String response = "";
    char c;
    while (client.connected() && (c = client.read()) != -1) {
      response += c;
    }

    client.stop();

    // Parse the payload
    parse_payload(response);
    } else {
    //Serial.println("Failed to connect to server.");
  }

  // Continue with the rest of your code...
}




//*************************************************************************
// Function: parse the payload which is the "Stations" text file retrieved
// from the web server.
//*************************************************************************
void parse_payload(String payload) {
  // Reset category and station indices
  categoryIndex = -1;
  stationIndex = 0;

  // Split payload into lines
  String line;
  int index = 0;
  while ((index = payload.indexOf('\n')) != -1) {
    line = payload.substring(0, index);
    payload = payload.substring(index + 1);

    // Parse each line
    if (line.startsWith("[")) {
      // Category line
      categoryIndex++;
      stations[categoryIndex][0] = ""; // Initialize first station of new category
      categories[categoryIndex] = line.substring(1, line.length() - 2);
      stationIndex = 0;
    } else {
      // Station line
      stations[categoryIndex][stationIndex] = line;
      stationIndex++;
    }
  }

  // Parse the last line
  if (!payload.isEmpty()) {
    if (payload.startsWith("[")) {
      // Category line
      categoryIndex++;
      stations[categoryIndex][0] = ""; // Initialize first station of new category
      categories[categoryIndex] = payload.substring(1, payload.length() - 1);
    } else {
      // Station line
      stations[categoryIndex][stationIndex] = payload;
    }
  }
}

//*************************************************************************
// Function to get the number of categories
//*************************************************************************
int getNumberOfCategories() {
  int count = 0;
  for (int i = 0; i < MAX_CATEGORIES; i++) {
    if (!categories[i].isEmpty()) {
      count++;
    }
  }
  return count;
}

//*************************************************************************
// Function to get the number of stations for a given category index
//*************************************************************************
int getNumberOfStations(int categoryIndex) {
  int count = 0;
  if (categoryIndex >= 0 && categoryIndex <= MAX_CATEGORIES) {
    for (int i = 0; i < MAX_STATIONS_PER_CATEGORY; i++) {
      if (!stations[categoryIndex][i].isEmpty()) {
        count++;
      }
    }
  }
  return count;
}
