#include "WiFiSetup.h"

AsyncWebServer server(80);

const char* PARAM_INPUT_1 = "input1";
const char* PARAM_INPUT_2 = "input2";
const char* PARAM_INPUT_3 = "input3";
const char* PARAM_INPUT_4 = "input4";

const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML>
<html>
<head>
  <title>iRadio WiFi Setup Form</title>
  <style>
    h1 {
      font-size: 50px;
    }
    p {
      font-size: 38px;
      text-align: center;
    }
    .form-container {
      display: flex;
      flex-direction: column;
      width: 80%;
      margin: auto;
    }
    input[type="text"] {
      font-size: 38px;
      padding: 8px;
      margin-bottom: 20px;
      width: 100%;
    }
    input[type="submit"] {
      font-size: 38px;
      padding: 8px 20px;
      margin-bottom: 20px;
      width: 100%;
    }
  </style>
</head>
<body>
  <h1><center>iRadio WiFi Setup Form</center></h1>
  <br>
  <p>Use the form to complete each field, one at a time. A confirmation will be displayed after each submission.</p>
  <div class="form-container">
    <form action="/get">
      <label for="input1"></label>
      <input type="text" id="input1" name="input1" placeholder="WiFi SSID">
      <br>
      <label for="input2"></label>
      <input type="text" id="input2" name="input2" placeholder="WiFi Password">
      <br>
      <label for="input3"></label>
      <input type="text" id="input3" name="input3" placeholder="Hostname">
      <br>
      <label for="input4"></label>
      <input type="text" id="input4" name="input4" placeholder="Path and File">
      <br>
      <input type="submit" value="Submit">
    </form>
  </div>
</body>
</html>)rawliteral";

void notFound(AsyncWebServerRequest *request) {
  request->send(404, "text/plain", "Not found");
}

String readFile(fs::FS &fs, const char * path){
  Serial.printf("Reading file: %s\r\n", path);
  File file = fs.open(path, "r");
  if(!file || file.isDirectory()){
    Serial.println("- empty file or failed to open file");
    return String();
  }
  Serial.println("- read from file:");
  String fileContent;
  while(file.available()){
    fileContent += String((char)file.read());
  }
  file.close();
  Serial.println(fileContent);
  return fileContent;
}

void writeFile(fs::FS &fs, const char * path, const char * message){
  Serial.printf("Writing file: %s\r\n", path);
  File file = fs.open(path, "w");
  if(!file){
    Serial.println("- failed to open file for writing");
    return;
  }
  if(file.print(message)){
    Serial.println("- file written");
  } else {
    Serial.println("- write failed");
  }
  file.close();
}

void setupWiFi() {
  if(!SPIFFS.begin(true)){
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }

  const char* ssid     = "iRadio_Setup";
  const char* password = "123456789";

  Serial.print("Setting AP (Access Point)…");
  WiFi.softAP(ssid, password);
  IPAddress IP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(IP);

  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html);
  });

  server.on("/restart", HTTP_GET, [] (AsyncWebServerRequest *request) {
  ESP.restart();
  });

  server.on("/get", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    String inputParam;
    String response = "<h1>The following values have been set:</h1><br>";

    if (request->hasParam(PARAM_INPUT_1)) {
        inputMessage = request->getParam(PARAM_INPUT_1)->value();
        inputParam = "WiFi SSID";
        writeFile(SPIFFS, "/ssid", inputMessage.c_str());
        response += "SSID: " + inputMessage + "<br>";
    }
    if (request->hasParam(PARAM_INPUT_2)) {
        inputMessage = request->getParam(PARAM_INPUT_2)->value();
        inputParam = "WiFi Password";
        writeFile(SPIFFS, "/password", inputMessage.c_str());
        response += "Password: " + inputMessage + "<br>";
    }
    if (request->hasParam(PARAM_INPUT_3)) {
        inputMessage = request->getParam(PARAM_INPUT_3)->value();
        inputParam = "Hostname";
        writeFile(SPIFFS, "/host", inputMessage.c_str());
        response += "Host: " + inputMessage + "<br>";
    }
    if (request->hasParam(PARAM_INPUT_4)) {
        inputMessage = request->getParam(PARAM_INPUT_4)->value();
        inputParam = "Path and File";
        writeFile(SPIFFS, "/path", inputMessage.c_str());
        response += "File and path: " + inputMessage + "<br>";
    }
    else {
        inputMessage = "No message sent";
        inputParam = "none";
    }

// Reboot instructions
response += "<p style=\"font-size: 32px;\">If these values are correct, please re-enter them.  If they are correct, press restart. ";
response += "You can now enjoy listening to the iRadio.</p>";

// Button to redo
response += "<form action=\"/\" method=\"get\">";
response += "<button type=\"submit\" style=\"font-size: 38px;\">Re-enter Values</button>";
response += "</form>";

// Button to restart
response += "<form action=\"/restart\" method=\"get\">";
response += "<button type=\"submit\" style=\"font-size: 38px;\">Restart radio</button>";
response += "</form>";

    request->send(200, "text/html", response);
  });

  server.onNotFound(notFound);
  server.begin();

  //While in Setup mode, we want to just stop here and have the user 
  //reboot and then the radio will boot normally.
  while(true) {}
}
